/* bbclist.c (Acorn BBC Basic file viewer)

   Copyright (C) W.H.Scholten 2002,2003,2009,2014,2016
  
   Permission to use, copy, modify, distribute, and sell this software
   and its documentation for any purpose is hereby granted without fee,
   provided that the above copyright notice appear in all copies and
   that both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of the copyright holder
   not be used in advertising or publicity pertaining to distribution
   of the software without specific, written prior permission. The
   copyright holder makes no representations about the suitability of
   this software for any purpose. It is provided "as is" without express
   or implied warranty.
  
   THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
   SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
   SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  
  
   Send comments and bug-reports to 
  
      swhs-web@xs4all.nl
*/


/*
 * History:
 * --------
 * Version 0.1-0.7: not released.
 * Version 0.8    20 Jan 2003
 * Version 0.9    11 Nov 2003
 * Version 0.10   26 Feb 2009
 * Version 0.11   26 Feb 2009
 * Version 0.12   28 Mar 2009 (not released)
 * Version 0.13   23 Apr 2009
 * Version 0.14    4 May 2009
 * Version 0.15   14 May 2009
 * Version 0.15.1 21 May 2009
 * Version 0.16   20 Feb 2014 (not released)
 * Version 0.17   13 Aug 2016
 * Version 0.18   15 Aug 2016
 *
 * 9 Dec. 2002: Experimented with bas2txt to handle nested DEF PROCs and
 *   colourised HTML output.
 *
 * 12 Dec. 2002: Decided to write a new program to properly handle assembler
 *   listings, in particular ORA (which became OR A). Then added colour html
 *   output, (probably?) correct inline line numbers etc.
 *
 * 11 Nov. 2003: Added a hexdump for any data at the end of the Basic program.
 *
 * 25 Feb. 2009: HTML change to HTML 4.01-strict + CSS styles (inline)
 *   I noticed TOP is actually TO+P, so colouring of TOP was incorrect, and
 *   pretty-printing of TOP was also incorrect (it gave " TO P").
 *   Both problems have been fixed.
 *
 * 26 Feb 2009: actually, decided to do it properly with span classes. Also,
 *   TOP should not be in the class of fornext, fixed in a somewhat messy way.
 *
 * 14 Mar 2009: Colourise control chars with their own colour, not that for
 *   'assembler comment'.
 *
 * 28 Mar 2009: The code for -n (no line numbers) is now actually there.
 *
 * 5 Apr 2009: In the hexdump, in case of HTML output, replace the ascii '<',
 *   '>', '&' by "&lt;", "&gt;", "&amp;".
 *   Options explained to be irreversible for simplicity.
 *
 * 4 May 2009: In the hexdump, correctly terminate the ascii printing for
 *   standard text output (added the line: ascii[i+1] = 0).
 *   Added a man page.
 *
 * 7 May 2009: web page reference in help output, colourised using definitions
 *   in _ansi_console.h
 *
 * 13 May 2009: In the bmore script, if the file is a directory, list that
 *   directory. The Makefile now also installs the bmore script. Because of shell
 *   restrictions, I'm including a call to 'cat' if a file doesn't turn out to be
 *   a Basic file. This means with 'bmore' you can view Basic files, other files,
 *   and directories. Very useful.
 *
 * 20 May 2009: Esp. some changes to the Makefile to create the man1 directory etc.
 *
 * 19,20 Feb 2014:
 *  c Reorganised the help/use code and output.
 *  + Added option to colourise the text output to the terminal.
 *  c Changed the output options to -c|-t|-h for output as ANSI colour text,
 *    (default), plain text, html.
 *  c Made output to stdout the default and added option -f to send the output
 *    to file. Output filename for -c is <name>.ansicolour.txt
 *  c removed option -O (which selected output to be sent to stdout)
 *  c Manual page updated for these changes.
 *
 *
 * 20,21 Mar 2014:
 *  c Options become long (readable!), but single character options are
 *    retained as abbreviations. Options are no longer irreversible as when making
 *    the options longer then conceptually it's better to make cases available even
 *    if they are not used nor needed, simply to have a logical structure :)
 *  c Don't check for unterminated strings in assembler- nor Basic comments.

 * 13-8-2016:
 *  c small changes to the man page and more.
 *
 * 15-8-2016:
 *  + added the short option '-o' (it was in the code, but commented out)
 *  c added error message about missing file if options were given without a filename.
 *  c don't close 'fp_uit' if it is stdout (multiple files in one go works now to stdout).
 *  c free allocated memory 'file_data'.
 *
 *
 * Possible improvements:
 * ----------------------
 *  - Check if ENDPROCs are after an IF, if so, do not decrease indent
 *     level
 *
 *  - Indent FNs, and as with ENDPROC, check whether an end-of-function
 *     is conditional.
 *
 *  - Add a switch to replace embedded ctrl chars by spc to produce a
 *     directly spoolable file?
 *
 *  - Check for bad Basic programs, i.e. the line length byte should be
 *     OK.
 *
 *
 * Old HTML output:
 * ----------------
 * If you'd like the HTML output code to use FONT (deprecated in HTML4, i.e.
 * since ca. 1999, but since about 2001 CSS started to be used in earnest),
 * then replace the following in this program:
 *
 * <span class=\"basic_comment\">    by      <FONT COLOR=\""COLOUR__BASIC_COMMENT"\">
 * </span>                           by      </FONT>
 * etc.
 */



#include <stdio.h>
#include <string.h>
#include <stdlib.h>
//#include <ctype.h>

#include "_config.h"

/* For execvp */
#include <unistd.h>

/* Console output: */
#define USE_ANSI_COLOURS 1

#include "_ansi_console.h"


#define COLOUR__BASIC_COMMENT "#777777" /* After REM */
#define COLOUR__ASSEM_COMMENT "#777777" /* After \ */
#define COLOUR__STRUCT        "#880000" /* DEF, PROC, FN, ENDPROC */
#define COLOUR__FORNEXT       "#880088" /* FOR TO STEP NEXT */
#define COLOUR__LOOP          "#880088" /* REPEAT UNTIL */
#define COLOUR__IFTHEN        "#008800" /* IF THEN ELSE */
#define COLOUR__LINENUMBER    "#008888" /* After (ON) GOTO, (ON) GOSUB, THEN  */
#define COLOUR__CONTROL_CHARS "#004488" /* embedded control chars */

#define COLOUR__CONSOLE__BASIC_COMMENT CONSOLE_GRAY    /* After REM */
#define COLOUR__CONSOLE__ASSEM_COMMENT CONSOLE_GRAY    /* After \ */
#define COLOUR__CONSOLE__STRUCT        CONSOLE_RED     /* DEF, PROC, FN, ENDPROC */
#define COLOUR__CONSOLE__FORNEXT       CONSOLE_MAGENTA /* FOR TO STEP NEXT */
#define COLOUR__CONSOLE__LOOP          CONSOLE_MAGENTA /* REPEAT UNTIL */
#define COLOUR__CONSOLE__IFTHEN        CONSOLE_GREEN   /* IF THEN ELSE */
#define COLOUR__CONSOLE__LINENUMBER    CONSOLE_CYAN    /* After (ON) GOTO, (ON) GOSUB, THEN  */
#define COLOUR__CONSOLE__CONTROL_CHARS CONSOLE_BLUE    /* embedded control chars */
#define COLOUR__CONSOLE__BOLD          CONSOLE_BOLD
#define COLOUR__CONSOLE__RESET         CONSOLE_RESTORE


/* Basic 1 token definitions, &80 - &FF */
const char BASIC_1_TOKENS[128][10] = {
    /*         0 / 8       1 / 9        2 / A        3 / B        4 / C        5 / D        6 / E        7 / F       */
    /* 0x80 */ "AND",      "DIV",       "EOR",       "MOD",       "OR",        "ERROR",     "LINE",      "OFF",
    /* 0x88 */ "STEP",     "SPC",       "TAB(",      "ELSE",      "THEN",      " x8D ",     " x8E ",     "PTR",
    /* 0x90 */ "PAGE",     "TIME",      "LOMEM",     "HIMEM",     "ABS",       "ACS",       "ADVAL",     "ASC",
    /* 0x98 */ "ASN",      "ATN",       "BGET",      "COS",       "COUNT",     "DEG",       "ERL",       "ERR",
    /* 0xA0 */ "EVAL",     "EXP",       "EXT",       "FALSE",     "FN",        "GET",       "INKEY",     "INSTR(",
    /* 0xA8 */ "INT",      "LEN",       "LN",        "LOG",       "NOT",       "OPENIN",    "OPENOUT",   "PI",
    /* 0xB0 */ "POINT(",   "POS",       "RAD",       "RND",       "SGN",       "SIN",       "SQR",       "TAN",
    /* 0xB8 */ "TO",       "TRUE",      "USR",       "VAL",       "VPOS",      "CHR$",      "GET$",      "INKEY$",
    /* 0xC0 */ "LEFT$(",   "MID$(",     "RIGHT$(",   "STR$",      "STRING$(",  "EOF",       "AUTO",      "DELETE",
    /* 0xC8 */ "LOAD",     "LIST",      "NEW",       "OLD",       "RENUMBER",  "SAVE",      " xCE ",     "PTR",
    /* 0xD0 */ "PAGE",     "TIME",      "LOMEM",     "HIMEM",     "SOUND",     "BPUT",      "CALL",      "CHAIN",
    /* 0xD8 */ "CLEAR",    "CLOSE",     "CLG",       "CLS",       "DATA",      "DEF",       "DIM",       "DRAW",
    /* 0xE0 */ "END",      "ENDPROC",   "ENVELOPE",  "FOR",       "GOSUB",     "GOTO",      "GCOL",      "IF",
    /* 0xE8 */ "INPUT",    "LET",       "LOCAL",     "MODE",      "MOVE",      "NEXT",      "ON",        "VDU",
    /* 0xF0 */ "PLOT",     "PRINT",     "PROC",      "READ",      "REM",       "REPEAT",    "REPORT",    "RESTORE",
    /* 0xF8 */ "RETURN",   "RUN",       "STOP",      "COLOUR",    "TRACE",     "UNTIL",     "WIDTH",     " xFF "
};


/* token_space: how to pretty-print the token:
 * 0 = no space before or after the token
 * 1 = pre space (not for functions)
 * 2 = post space (not for tokens with implicit bracket)
 * 3 = pre + post space
 */
int token_space[128] = {
	/*          0/8   1/9   2/A   3/B   4/C   5/D   6/E   7/F */
	/* 0x80 */    3,    3,    3,    3,    3,    3,    3,    3,
	/* 0x88 */    3,    3,    1,    3,    3,    3,    2,    3,
	/* 0x90 */    0,    0,    0,    0,    2,    2,    2,    2,
	/* 0x98 */    2,    2,    2,    2,    2,    2,    2,    2,
	/* 0xA0 */    2,    2,    2,    2,    0,    2,    2,    1,
	/* 0xA8 */    2,    2,    2,    2,    2,    2,    2,    2,
	/* 0xB0 */    1,    2,    2,    2,    2,    2,    2,    2,
	/* 0xB8 */    3,    2,    2,    2,    2,    2,    2,    2,
	/* 0xC0 */    1,    1,    1,    2,    1,    3,    3,    3,
	/* 0xC8 */    3,    3,    3,    3,    3,    3,    3,    2,
	/* 0xD0 */    0,    0,    0,    0,    3,    3,    3,    3,
	/* 0xD8 */    3,    3,    3,    3,    3,    3,    3,    3,
	/* 0xE0 */    3,    3,    3,    3,    3,    3,    3,    3,
	/* 0xE8 */    3,    3,    3,    3,    3,    3,    3,    3,
	/* 0xF0 */    3,    3,    1,    3,    3,    3,    3,    3,
	/* 0xF8 */    3,    3,    3,    3,    3,    3,    3,    3,
};


/* Changes in the Basic 2 token definitions &80 - &FF as compared to those in
 * Basic 1:
 *
 * Token | Explanation
 * ------+---------------
 * &8E   | OPENIN, has a different function to the one in Basic 1; Files are
 *       |   opened read-only in Basic 2 with this function.
 * &AD   | New keyword OPENUP (open a file for reading & writing); this has the
 *       |   same function as OPENIN in Basic 1 (and uses the same token &AD)
 * &FF   | New keyword OSCLI
 */

/* Other keywords:
 * - The assembler mnemonic ORA is actually the keyword OR followed by the
 *    letter A (without a space!).
 * - The keyword TOP is actually the keyword TO followed by the letter P
 *    (without a space!).
 *
 * These cases need extra code for colouring and pretty-printing.
 */

/* Prototypes: */
void show_version(void);
void show_version_copyright(void);
void show_synopsis(void);
void show_options(void);
void use(void);
void missing_argument(void);


#define BBCLIST_VERSION "0.18"
#define BBCLIST_COPYRIGHT "(C) W.H.Scholten 2002-2016"


/* Functions: */
void show_version() {
	printf("bbclist "BBCLIST_VERSION"\n");
}


void show_version_copyright() {
	printf("\nbbclist, version "BBCLIST_VERSION", "BBCLIST_COPYRIGHT" (" CONSOLE_GRAY "http://wouter.bbcmicro.net/"CONSOLE_RESTORE")\n"
	"BBC Basic file viewer. Converts BBC Basic 1 & 2 to HTML, ASCII text or ANSI colour text\n\n"
		);
}

void missing_argument() {
	printf(COLOUR__CONSOLE__BOLD"Missing argument\n"COLOUR__CONSOLE__RESET);
	use();
}

void use() {
	show_version_copyright();
	show_synopsis();
	printf(
		"For further information on using bbclist, type:\n"
		"  bbclist -help\n"
		"\n"
		);
//	show_options();
	exit(1);
}

void show_synopsis() {
	printf(
	"Synopsis: bbclist [options] <filename> ...\n"
	"\n"
    "Default output: text with colour, to the console, with line numbers, no pretty printing, and programs are assumed to be Basic 2.\n"
	"\n"
		);
}

void show_options() {
	printf(
    "Options:\n"
	" -help : This output, showing options and abbreviations in square brackets.\n"
	"         So option -c is equivalent to -format colour, etc.\n"
	"\n"
	" -format {colour | text | html} : Set output format:\n"
	"     colour: text with ANSI escape sequences for colour. [-c]\n"
	"     text : plain text. [-t]\n"
	"     html : HTML (4.01-strict with CSS colouring). [-h]\n"
	"\n"
	" -output {file | stdout} : Set where output goes:\n"
	"     file : Output to a file named <filename>.ansicolour.txt, or <filename>.txt or <filename>.html depending on 'format'. [-f]\n"
	"     stdout : Output to 'standard output'. This is the default at start of the program. [-o]\n"
	"\n"
	" -basic {no-line-numbers | pretty-print | 1 | 2} : Set Basic options:\n"
	"     no-line-numbers : Output without line numbers. [-n]\n"
	"     pretty-print : Add spaces and indent structures for readability. [-p]\n"
	"     1 : Use the tokens of Basic 1. [-1]\n"
	"     2 : Use the tokens of Basic 2. [-2]\n"
	"\n"
		);
	exit(1);
}


#define False  0
#define True   1

/* State: */
int assembler = False;
int assembler_comment = False;
int basic_comment = False;

int line_found_if = False;

unsigned char *file_data;
int ptr;
int line_no;

//int conditional_level = 0; /* For ENDPROC and end-of-function checks within an IF or ELSE statement for pretty-printing */
 
int indent_level = 0;
int token_post_space = False;

/* options: */
int use_html = False;
int use_colour = True; /* For text files or terminal with ANSI colour sequences */
int use_stdout = True;
int use_line_numbers = True;
int use_pretty_print = False;
int basic_version = 2;


/* Hexdump code for data following the Basic program. */
#include "hex_dump.c"


void convert_file(char *fn);

int main(int argc, char**argv) {
	int  i;

	if (argc == 1) use();
	for (i=1; i<argc; i++) {
		if (strlen(argv[i]) >= 2 && argv[i][0] == '-') {
			if (strlen(argv[i]) > 2) {
				if (!strcmp(argv[i]+1, "help")) {
					show_options();
				}
				else {
					if (!strcmp(argv[i]+1, "format")) {
						if (argc <= i+1) {
							fprintf(stderr, "Missing argument for 'format'\n");
							exit(1);
						}
						if (!strcmp(argv[i+1], "colour")) {
							use_colour = True;
							use_html = False;
						}
						else if (!strcmp(argv[i+1], "text")) {
							use_colour = False;
							use_html = False;
						}
						else if (!strcmp(argv[i+1], "html")) {
							use_colour = True;
							use_html = True;
						}
						else {
							fprintf(stderr, "Bad argument '%s' for 'format'\n", argv[i+1]);
							exit(1);
						}
						i++;
					}

					else if (!strcmp(argv[i]+1, "output")) {
						if (argc <= i+1) {
							fprintf(stderr, "Missing argument for 'output'\n");
							exit(1);
						}

						if (!strcmp(argv[i+1], "file")) {
							use_stdout = False;
						}
						else if (!strcmp(argv[i+1], "stdout")) {
							use_stdout = True;
						}
						else {
							fprintf(stderr, "Bad argument '%s' for 'output'\n", argv[i+1]);
							exit(1);
						}
						i++;
					}

					else if (!strcmp(argv[i]+1, "basic")) {
						if (argc <= i+1) {
							fprintf(stderr, "Missing argument for 'basic'\n");
							exit(1);
						}

						if (!strcmp(argv[i+1], "no-line-numbers")) {
							use_line_numbers = False;
						}
						else if (!strcmp(argv[i+1], "line-numbers")) {
							use_line_numbers = True;
						}
						else if (!strcmp(argv[i+1], "pretty-print")) {
							use_pretty_print = True;
						}
						else if (!strcmp(argv[i+1], "1")) {
							basic_version = 1;
						}
						else if (!strcmp(argv[i+1], "2")) {
							basic_version = 2;
						}
						else {
							fprintf(stderr, "Bad argument '%s' for 'basic'\n", argv[i+1]);
							exit(1);
						}
						i++;
					}

					else {
						fprintf(stderr, "Bad option '%s'\n", argv[i]);
						exit(1);
					}
					if (argc <= i+1) missing_argument(); /* missing file */
				}

			} else { /* short options */
				switch(argv[i][1]) {
					case 'c':
						use_colour = True;
						use_html = False;
						break;

					case 't':
						use_colour = False;
						use_html = False;
						break;

					case 'h':
						use_colour = True;
						use_html = True;
						break;

					case 'f':
						use_stdout = False;
						break;

					case 'o':
						use_stdout = True;
						break;

					case 'n':
						use_line_numbers = False;
						break;

					case 'p':
						use_pretty_print = True;
						break;

					case '1':
						basic_version = 1;
						break;
					case '2':
						basic_version = 2;
						break;

					default:
						fprintf(stderr, "bad option '%c'\n", argv[i][1]);
						exit(1);
						break;
					}
				if (argc <= i+1) missing_argument(); /* missing file */
				}
		}
		else {
			convert_file(argv[i]);
		}

//		fprintf(stderr, "i=%d\n", i);
	}
	return 0;
}


/* Prototypes: */
int output_char(char *line, int ch);
int get_line_number(unsigned char *file_data, int file_len, int ptr);
int print_keyword(char *line, int token, int last_action_byte, int basic_version);
int last_char_space(char *line);
int last_char_space_colon(char *line);



int last_char_space(char *line) {
	if (strlen(line) == 0) return -1;

	return line[ strlen(line)-1 ] == ' ';
}


int last_char_space_colon(char *line) {
	if (strlen(line) == 0) return -1;

	return (line[ strlen(line)-1 ] == ' ' || line[ strlen(line)-1 ] == ':');
}


void convert_file(char *fn) {
	FILE *fp_in, *fp_uit;
	char fn_uit[255];
	int  file_len;

	int  st;
	int  quote_on = False;
	unsigned char curbyte;
	char line[1024];
	int  last_action_byte = -1;
	int  line_bytes;

	/* initialise state: */
//	assembler = False;
//	assembler_comment = False;
//	basic_comment = False;
//	line_found_if = False;
//	indent_level = 0;
//	token_post_space = False;


//	fprintf(stderr, "convert_file %s\n", fn);
	if (strlen(fn) > 250) {
		fprintf(stderr, "Filename %s too long\n", fn);
		exit(1);
	}

	fp_in = fopen(fn, "rb");
	if (fp_in == NULL) {
		fprintf(stderr, "Can't open file %s for input\n", fn);
		exit(1);
	}

	st = fseek(fp_in, 0L, SEEK_END);
	file_len = ftell(fp_in);
//fprintf(stderr, "file leng=%d\n", file_len);

	file_data = malloc(file_len);
	if (file_data == NULL) {
		fprintf(stderr, "Can't allocate enough memory\n");
		exit(1);
	}

	st = fseek(fp_in, 0L, SEEK_SET);
	st = fread(file_data, 1, file_len, fp_in);
	fclose(fp_in);
	if (st != file_len) {
		fprintf(stderr, "Error reading file %s\n", fn);
		exit(1);
	}

	if (file_data[0] != 13) {
		fprintf(stderr, CONSOLE_INVERT"Not a BBC Basic file, doesn't start with CR (&0D)"CONSOLE_RESTORE"\n");
#if EXIT_ON_NONBASIC_FILE
		exit(2);
#else
		int i, j;

		/* The following is because of shell problems (exit status in case of a string of commands), sh/bash etc. suck pretty hard */
		/* XXX Should make this an ASCII or UTF-8 check... */
		for (i=0; i<32; i++) {
			j=j | file_data[i];
		}
		if ((i &128)==0) {
			/* Looks like ascii, so do 'more' */
			char *m_argv[3];
			m_argv[0] = "cat";
			m_argv[1] = fn;
			m_argv[2] = NULL;
			execvp("cat", m_argv);
		}
#endif
	}

	if (use_stdout) {
		fp_uit = stdout;
	}
	else {
		strcpy(fn_uit, fn);
		if (use_html)
			strcat(fn_uit, ".html");
		else if (use_colour)
			strcat(fn_uit, ".ansicolour.txt");
		else
			strcat(fn_uit, ".txt");

		fp_uit = fopen(fn_uit, "wb");
		if (fp_uit == NULL) {
			fprintf(stderr, "Can't open file %s for output\n", fn_uit);
			exit(1);
		}
	}

	if (use_html) {
		/* HTML start */
		fprintf(fp_uit, "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\n");
		fprintf(fp_uit, "<html>\n");
		fprintf(fp_uit, "  <head>\n");
		fprintf(fp_uit, "    <meta http-equiv=\"Content-Type\" content=\"text/html;charset=utf-8\">\n");

		fprintf(fp_uit, "    <style type=\"text/css\">\n");
		fprintf(fp_uit, "    h1 {color: #221122}\n");
		fprintf(fp_uit, "    h2 {color: #113355}\n");
		fprintf(fp_uit, "    body {font-family: arial, \"lucida console\", sans-serif }\n");
		fprintf(fp_uit, "    span.basic_comment { color:"COLOUR__BASIC_COMMENT" }\n");
		fprintf(fp_uit, "    span.assem_comment { color:"COLOUR__ASSEM_COMMENT" }\n");
		fprintf(fp_uit, "    span.struct { color:"COLOUR__STRUCT" }\n");
		fprintf(fp_uit, "    span.fornext { color:"COLOUR__FORNEXT" }\n");
		fprintf(fp_uit, "    span.loop { color:"COLOUR__LOOP" }\n");
		fprintf(fp_uit, "    span.ifthen { color:"COLOUR__IFTHEN" }\n");
		fprintf(fp_uit, "    span.linenumber { color:"COLOUR__LINENUMBER" }\n");
		fprintf(fp_uit, "    span.control_chars { color:"COLOUR__CONTROL_CHARS" }\n");
		fprintf(fp_uit, "    </style>\n");

		fprintf(fp_uit, "    <meta name=\"generator\" content=\"bbclist "BBCLIST_VERSION" by W.H.Scholten\">\n");

		fprintf(fp_uit, "    <title>BBC basic program \"%s\"</title>\n", fn);
		fprintf(fp_uit, "  </head>\n" "  <body>\n\n" "<pre>\n");
	}

	ptr = -1;
	line_no = -1;
	quote_on = False;
	line[0] = 0; /* empty line */
	last_action_byte = -1;
	line_bytes = 0;
	for (;;) {
		ptr++;
		if (ptr >= file_len) break;

		curbyte = file_data[ ptr ];
		line_bytes++;

		if (strlen(line) > 512 || line_bytes >= 256) {
			/* 2nd case is a garbage Basic section, so output what we have, add a line feed for visuals, and continue */
			if (line_bytes >= 256) { strcat(line, "\n"); line_bytes = 0; }
			fwrite(line, 1, strlen(line), fp_uit); /* No fprintf as % may appear in the line */
			line[0] = 0;
		}

		/* New line */
		if (curbyte == 13) {
			int  i;

			line_bytes = 0;

			/* Print previous line */
			if (assembler && assembler_comment) {
				if (use_html) {
					strcat(line, "</span>"); /* Colour off */
				}
				else if (use_colour) {
					strcat(line, COLOUR__CONSOLE__RESET); /* Colour off */
				}
//				assembler_comment = False;
			}
			if (!assembler && basic_comment) {
				if (use_html) {
					strcat(line, "</span>"); /* Colour off */
				}
				else if (use_colour) {
					strcat(line, COLOUR__CONSOLE__RESET); /* Colour off */
				}
//				basic_comment = False;
			}
			strcat(line, "\n");
			fwrite(line, 1, strlen(line), fp_uit); /* No fprintf as % may appear in the line */
			line[0] = 0;

//			if (quote_on) {
			if (quote_on && !assembler_comment && !basic_comment) {
				fprintf(stderr, "Unterminated string in Basic linenumber %d\n", line_no);
			}
			quote_on = False;
			assembler_comment = False;
			basic_comment = False;
			last_action_byte = 13;

			ptr++;
			line_no = get_line_number(file_data, file_len, ptr);
			if (line_no < 0) break; /* EOF */
			if (line_no == 32768) break; /* End of Basic program */

//Debug			printf("line_no = %5d, ptr=%d\n", line_no, ptr);
			ptr += 2; /* also skips the line-length byte after the linenumber */
			if (use_line_numbers)
				sprintf(line, "%5d", line_no);
			else
				strcpy(line, "");

			if (use_pretty_print) {
				strcat(line, " ");
				for (i=0; i<indent_level; i++) strcat(line, " ");
			}
			continue;
		}


		if (quote_on) {
			output_char(line, curbyte);

			if (curbyte == '"') {
				quote_on = !quote_on;
			}
			continue;
		}

		if (curbyte == ' ') {
			if (!use_pretty_print || assembler || basic_comment || (last_action_byte < 0x80 /* not following a token, so space must be kept */)) strcat(line, " ");
		}
		else if (curbyte == '"') {
			quote_on = True;
			strcat(line, "\"");
			last_action_byte = '"';
		}

		else if (curbyte == 'P' && last_action_byte == 0xB8 /* TO P */) {
			/* This section is to fix formatting/colouring problems caused by TOP being the keyword TO + letter P */
			if (use_html) {
				/* First remove the keyword 'TO' and the colour */
				int replace_offset = strlen(line) - strlen("<span class=\"fornext\">TO</span></b>");
				strcpy(line+replace_offset, "TO");
				output_char(line, curbyte);
				strcat(line, "</b>");
			}
			else if (use_colour) {
				/* First remove the keyword 'TO' and the colour */
				int replace_offset = strlen(line) - strlen(COLOUR__CONSOLE__FORNEXT "TO" COLOUR__CONSOLE__RESET COLOUR__CONSOLE__RESET);
//				fprintf(stderr, "offset=%d (line from that point is: %s)\n", replace_offset, line+replace_offset);
				strcpy(line+replace_offset, "TO");
				output_char(line, curbyte);
				strcat(line, COLOUR__CONSOLE__RESET); /* Bold off (also turns Colour off) */

			}
			else {
				output_char(line, curbyte);
			}
			last_action_byte = curbyte;
		}

		else if (curbyte >= 0x80) {
			switch(curbyte) {
//			case 0xE7: /* IF */
//				conditional_level++;
//				conditional_type = COND_IF;
//				break;
//			case 0x8B: /* ELSE */
//				conditional_type = COND_ELSE;
//				break;

			case 0xA4: /* FN */
				if (last_action_byte == 0xDD) indent_level--; break;
			case 0xF2: /* PROC */
				break;
			case 0xDD: /* DEF */
				indent_level++; break;
			case 0xE3: /* FOR */
			case 0xF5: /* REPEAT */
				indent_level++; break;
			case 0xE1: /* ENDPROC, setting to 0 takes care of multiple DEF PROC entry points with single ENDPROC */
				indent_level = 0; break;
			case 0xED: /* NEXT */
			case 0xFD: /* UNTIL */
				indent_level--; break;
			default: break;
			}

			if (use_pretty_print && (last_action_byte >= 0x80) && token_post_space) {
				strcat(line, " ");
			}
			print_keyword(line, curbyte, last_action_byte, basic_version);

			last_action_byte = curbyte;
		}
		else { /* 0-127, except ' ', '"' */
			if (assembler && curbyte == ']') { /* End of assembly code */
				output_char(line, curbyte);
				assembler = False;
			}
			else if (assembler && curbyte == '\\') { /* Assembly comment */
				if (use_html) {
					strcat(line, "<span class=\"assem_comment\">");
				}
				else if (use_colour) {
					strcat(line, COLOUR__CONSOLE__ASSEM_COMMENT); /* assembler comment */
				}

				output_char(line, curbyte);
				assembler_comment = True;
			}
			else if (assembler && curbyte == ':') { /* ':' ends the comment in the line, the rest is interpreted as assembly code... */
				if (assembler_comment && use_html) {
					strcat(line, "</span>"); /* Colour off */
				}
				else if (assembler_comment && use_colour) {
					strcat(line, COLOUR__CONSOLE__RESET); /* Colour off */
				}

				output_char(line, curbyte);
				assembler_comment = False;
			}
			else if ((last_action_byte == 13 || last_action_byte == ':') && curbyte == '[') {
				output_char(line, curbyte);
				assembler = True;
			}
			else {
				if (use_pretty_print && last_action_byte >= 0x80 && token_post_space) {
					if (((curbyte != ':' && curbyte != 13 && curbyte != '(') || (curbyte == '(' && (token_space[ last_action_byte & 127 ] == 3)))) {
						strcat(line, " ");
					}
				}

				output_char(line, curbyte);
			}
			last_action_byte = curbyte;
		}
	}

//Debug	fprintf(stderr, "END: line_no=%d, file_len=%d, ptr=%d\n", line_no, file_len, ptr);

	if(strlen(line) > 0) {
		fprintf(stderr, "Bad Basic file: the last line ended without a CR\n");
		fwrite(line, 1, strlen(line), fp_uit); /* No fprintf as % may appear in line */
	}

	if (quote_on) {
		fprintf(stderr, "Unterminated string in Basic linenumber %d\n", line_no);
	}


	fprintf(fp_uit, "\n");
	hexdmp(fp_uit, file_data, file_len, ptr); /* hexdump any left over bytes... */

	if (use_html) {
		/* HTML end */
		fprintf(fp_uit, "</pre>\n" "\n  </body>\n" "</html>\n");
	}

	free(file_data);
	if (!use_stdout) fclose(fp_uit);
}


int output_char(char *line, int ch) {
	int j = strlen(line);

	if (use_html) {
		switch (ch) {
		case '<': sprintf(&line[ j ], "&lt;"); break;
		case '>': sprintf(&line[ j ], "&gt;"); break;
		case '&': sprintf(&line[ j ], "&amp;"); break;
		default:
			if (ch < 32 || ch > 126) {
				/* control characters */
				strcat(line, "<span class=\"control_chars\">");
				j = strlen(line);
				sprintf(&line[ j ], "[0x%02x]", ch);
				strcat(line, "</span>"); /* Colour off */
			}
			else {
				line[ j ] = ch;
				line[ j+1 ] = 0;
			}
			break;
		}
	}
	else {
		if (ch < 32 || ch > 126) {
			/* control characters */
			if (use_colour) {
				/* Use COLOUR__CONSOLE__RESET first for bold that is used for the comments */
				sprintf(&line[ j ], COLOUR__CONSOLE__RESET COLOUR__CONSOLE__CONTROL_CHARS"[0x%02x]"COLOUR__CONSOLE__RESET, ch);
				/* html spans can be nested, but consoled reset resets everything so return to previous status
				 * which is bold black = gray for comments (should save status and depending on that fix it) */
				if (assembler_comment)
					strcat(line, COLOUR__CONSOLE__ASSEM_COMMENT);
				else if (basic_comment)
					strcat(line, COLOUR__CONSOLE__BASIC_COMMENT);
			}
			else {
				sprintf(&line[ j ], "[0x%02x]", ch);
			}
		}
		else {
			line[ j ] = ch;
			line[ j+1 ] = 0;
		}
	}

	if (j+1 > 1000) { /* This should never occur, see other check for output length */
		fprintf(stderr, "program error in bbclist: line too long\n");
		return -1;
	}
	return 0;
}


int get_line_number(unsigned char *file_data, int file_len, int ptr) {
	int line_no;

	if (ptr+1 > file_len+1) return -1; /* EOF */
	if (((int)file_data[ ptr ]) > 127) return 32768; /* End of Basic program */
	if (ptr+2 > file_len+1) return -1; /* EOF */
	line_no = ((int) file_data[ ptr+1 ]) | (((int)file_data[ ptr ]) << 8);

	return line_no;
}


int print_keyword(char *line, int token, int last_action_byte, int basic_version) {
	int colour_used = 0;
	int implicit_bracket = 0;

	int assembler_opcode = False;

    if (assembler && (last_action_byte == 13 || last_action_byte == ':')) {
		assembler_opcode = True;
	}
	else if (use_pretty_print && (token_space[token & 127] & 1) && !last_char_space_colon(line)) strcat(line, " ");
	token_post_space = False;

	if (!assembler && token == 0xF4) basic_comment = True;

	if ((use_html || use_colour) && !assembler_opcode && !assembler_comment) {
		/* Bold on */
		if (use_html)
			strcat(line, "<b>");
		else
			strcat(line, COLOUR__CONSOLE__BOLD);

		/* Colour ON */
		switch (token) {
		case 0xF2: /* PROC */
		case 0xA4: /* FN */
		case 0xDD: /* DEF */
		case 0xE1: /* ENDPROC */
			//case 0xA4: /* FN */
			if (use_html)
				strcat(line, "<span class=\"struct\">");
			else
				strcat(line, COLOUR__CONSOLE__STRUCT);
			colour_used = 1;
			break;
		case 0xE3: /* FOR */
		case 0xB8: /* TO */
		case 0x88: /* STEP */
		case 0xED: /* NEXT */
			if (use_html)
				strcat(line, "<span class=\"fornext\">");
			else
				strcat(line, COLOUR__CONSOLE__FORNEXT);
			colour_used = 1;
			break;
		case 0xF5: /* REPEAT */
		case 0xFD: /* UNTIL */
			if (use_html)
				strcat(line, "<span class=\"loop\">");
			else
				strcat(line, COLOUR__CONSOLE__LOOP);
			colour_used = 1;
			break;
		case 0xE7: /* IF */
		case 0x8C: /* THEN */
		case 0x8B: /* ELSE */
			if (use_html)
				strcat(line, "<span class=\"ifthen\">");
			else
				strcat(line, COLOUR__CONSOLE__IFTHEN);
			colour_used = 1;
			break;
		case 0x8D: /* Inline line number */
			if (use_html)
				strcat(line, "<span class=\"linenumber\">");
			else
				strcat(line, COLOUR__CONSOLE__LINENUMBER);
			colour_used = 1;
			break;
		default:
			break;
		}
	}

	if (token == 0x8D) { /* Inline line number */
		int n1, n2, n3;
		int inline_line_no;

		ptr++;
		n1 = file_data [ptr++];
		n2 = file_data [ptr++];
		n3 = file_data [ptr];

		inline_line_no = (n3 - 0x40) * 256 + (n2 - 0x40);
		switch(n1) {
		case 0x44: inline_line_no += 0x40; break;
		case 0x54: inline_line_no += 0x00; break;
		case 0x64: inline_line_no += 0xC0; break;
		case 0x74: inline_line_no += 0x80; break;

		case 0x40: inline_line_no += 0x4040; break;
		case 0x50: inline_line_no += 0x4000; break;
		case 0x60: inline_line_no += 0x40C0; break;
		case 0x70: inline_line_no += 0x4080; break;

		default:
			fprintf(stderr, "Bad inline line number at byte #%d (line %d), starting with byte: 0x%02X\n", ptr-2, line_no, n1);
			break;
		}

		int j = strlen (line);
		sprintf(&line[j], "%d", inline_line_no);
	}
	else {
		switch (basic_version) {
		case 1: {
			const char *token_str = BASIC_1_TOKENS[token-128];
			char token_str2[10];
			int off;

			strcpy(token_str2, token_str);

			off = strlen(token_str2)-1;

			if (token_str2[ off ] == '(') {
				token_str2[ off ] = 0;
				implicit_bracket = 1;
//fprintf(stderr, "test 1\n");
			}
			strcat(line, token_str2);
			if (token_str2[0] == ' ')
				fprintf(stderr, "Bad opcode 0x%02x at byte #%d (line %d)\n", token, ptr, line_no);
			break;
		}

		case 2: {
			const char *token_str = BASIC_1_TOKENS[token-128];
			char token_str2[10];
			int off;

			if (token == 0x8E) token_str = "OPENIN";
			if (token == 0xAD) token_str = "OPENUP";
			if (token == 0xFF) token_str = "OSCLI";

			strcpy(token_str2, token_str);

			off = strlen(token_str2)-1;
			if (token_str2[ off ] == '(') {
				token_str2[ off ] = 0;
				implicit_bracket = 1;
//fprintf(stderr, "test 2, token 0x%02x, Basic line %d\n", token, linenumber);
			}
			strcat(line, token_str2);
			if (token_str2[0] == ' ')
				fprintf(stderr, "Bad opcode 0x%02x at byte #%d (line %d)\n", token, ptr, line_no);
			break;
		}

		default:
			fprintf(stderr, "Bad Basic version %d\n", basic_version);
			exit(1);
			break;
		} /* switch */
	}


	if ((use_html || use_colour) && !assembler_opcode && !assembler_comment) {
		/* Colour off */
		if (colour_used) {
			if (use_html)
				strcat(line, "</span>"); /* Colour off */
			else /* if (use_colour)*/ {
				strcat(line, COLOUR__CONSOLE__RESET); /* Colour off which also restores bold if it was used. */
			}
		}
		/* Bold off */
		if (use_html)
			strcat(line, "</b>");
		else if (use_colour)
			strcat(line, COLOUR__CONSOLE__RESET); /* Colour off which also restores bold if it was used. */
	}

    if (assembler_opcode) {
	}
	else {
		if (use_pretty_print && (token_space[token & 127] & 2)) token_post_space = True;
		if (token == 0xF4) {
			if (use_html)
				strcat(line, "<span class=\"basic_comment\">"); /* REM */
			else if (use_colour)
				strcat(line, COLOUR__CONSOLE__BASIC_COMMENT); /* REM */

			basic_comment = True;
		}
	}

	if (implicit_bracket) strcat(line, "(");

	return 0;
}
